@@ -21,12 +21,11 @@ import tarfile
 from threading import Event
 
 from ipython_genutils.tempdir import TemporaryDirectory
-from ipython_genutils.py3compat import which
 from jupyter_core.paths import jupyter_config_path
 from notebook.nbextensions import GREEN_ENABLED, GREEN_OK, RED_DISABLED, RED_X
 
 from .semver import Range, gte, lt, lte, gt
-from .jlpmapp import YARN_PATH, HERE
+from .jlpmapp import YARN_PATH, HERE, which
 from .process import Process, WatchHelper
 
 
@@ -44,7 +43,7 @@ def pjoin(*args):
 
 
 def get_user_settings_dir():
-    """Get the configured JupyterLab app directory.
+    """Get the configured JupyterLab user settings directory.
     """
     settings_dir = os.environ.get('JUPYTERLAB_SETTINGS_DIR')
     settings_dir = settings_dir or pjoin(
@@ -53,6 +52,16 @@ def get_user_settings_dir():
     return osp.realpath(settings_dir)
 
 
+def get_workspaces_dir():
+    """Get the configured JupyterLab workspaces directory.
+    """
+    workspaces_dir = os.environ.get('JUPYTERLAB_WORKSPACES_DIR')
+    workspaces_dir = workspaces_dir or pjoin(
+        jupyter_config_path()[0], 'lab', 'workspaces'
+    )
+    return osp.realpath(workspaces_dir)
+
+
 def get_app_dir():
     """Get the configured JupyterLab app directory.
     """
@@ -89,14 +98,33 @@ def ensure_dev(logger=None):
         yarn_proc = Process(['node', YARN_PATH], cwd=parent, logger=logger)
         yarn_proc.wait()
 
-    if not osp.exists(pjoin(parent, 'dev_mode', 'build')):
+    if not osp.exists(pjoin(parent, 'dev_mode', 'static')):
         yarn_proc = Process(['node', YARN_PATH, 'build'], cwd=parent,
                             logger=logger)
         yarn_proc.wait()
 
 
-def watch_dev(logger=None):
-    """Run watch mode in a given directory.
+def ensure_core(logger=None):
+    """Ensure that the core assets are available.
+    """
+    staging = pjoin(HERE, 'staging')
+
+    # Bail if the static directory already exists.
+    if osp.exists(pjoin(HERE, 'static')):
+        return
+
+    if not osp.exists(pjoin(staging, 'node_modules')):
+        yarn_proc = Process(['node', YARN_PATH], cwd=staging, logger=logger)
+        yarn_proc.wait()
+
+    if not osp.exists(pjoin(HERE, 'static')):
+        yarn_proc = Process(['node', YARN_PATH, 'build'], cwd=staging,
+                            logger=logger)
+        yarn_proc.wait()
+
+
+def watch_packages(logger=None):
+    """Run watch mode for the source packages.
 
     Parameters
     ----------
@@ -126,12 +154,31 @@ def watch_dev(logger=None):
     tsf_proc = WatchHelper(['node', YARN_PATH, 'run', 'watch:files'],
         cwd=ts_dir, logger=logger, startup_regex=tsf_regex)
 
+    return [ts_proc, tsf_proc]
+
+
+def watch_dev(logger=None):
+    """Run watch mode in a given directory.
+
+    Parameters
+    ----------
+    logger: :class:`~logger.Logger`, optional
+        The logger instance.
+
+    Returns
+    -------
+    A list of `WatchHelper` objects.
+    """
+    logger = logger or logging.getLogger('jupyterlab')
+
+    package_procs = watch_packages(logger)
+
     # Run webpack watch and wait for compilation.
     wp_proc = WatchHelper(['node', YARN_PATH, 'run', 'watch'],
         cwd=DEV_DIR, logger=logger,
         startup_regex=WEBPACK_EXPECT)
 
-    return [ts_proc, tsf_proc, wp_proc]
+    return package_procs + [wp_proc]
 
 
 def watch(app_dir=None, logger=None):
@@ -181,13 +228,13 @@ def clean(app_dir=None):
             shutil.rmtree(target)
 
 
-def build(app_dir=None, name=None, version=None, logger=None,
-        command='build:prod', kill_event=None,
+def build(app_dir=None, name=None, version=None, public_url=None,
+        logger=None, command='build:prod', kill_event=None,
         clean_staging=False):
     """Build the JupyterLab application.
     """
     handler = _AppHandler(app_dir, logger, kill_event=kill_event)
-    return handler.build(name=name, version=version,
+    return handler.build(name=name, version=version, public_url=public_url,
                   command=command, clean_staging=clean_staging)
 
 
@@ -212,6 +259,13 @@ def disable_extension(extension, app_dir=None, logger=None):
     return handler.toggle_extension(extension, True)
 
 
+def check_extension(extension, app_dir=None, installed=False, logger=None):
+    """Check if a JupyterLab extension is enabled or disabled.
+    """
+    handler = _AppHandler(app_dir, logger)
+    return handler.check_extension(extension, installed)
+
+
 def build_check(app_dir=None, logger=None):
     """Determine whether JupyterLab should be built.
 
@@ -241,9 +295,11 @@ def unlink_package(package, app_dir=None, logger=None):
     return handler.unlink_package(package)
 
 
-def get_app_version():
+def get_app_version(app_dir=None):
     """Get the application version."""
-    return _get_core_data()['jupyterlab']['version']
+    app_dir = app_dir or get_app_dir()
+    handler = _AppHandler(app_dir)
+    return handler.info['version']
 
 
 # ----------------------------------------------------------------------
@@ -254,8 +310,6 @@ def get_app_version():
 class _AppHandler(object):
 
     def __init__(self, app_dir, logger=None, kill_event=None):
-        if app_dir and app_dir.startswith(HERE):
-            raise ValueError('Cannot run lab extension commands in core app')
         self.app_dir = app_dir or get_app_dir()
         self.sys_dir = get_app_dir()
         self.logger = logger or logging.getLogger('jupyterlab')
@@ -302,15 +356,18 @@ class _AppHandler(object):
             if other['path'] != info['path'] and other['location'] == 'app':
                 os.remove(other['path'])
 
-    def build(self, name=None, version=None, command='build:prod',
-            clean_staging=False):
+        return True
+
+    def build(self, name=None, version=None, public_url=None,
+            command='build:prod', clean_staging=False):
         """Build the application.
         """
         # Set up the build directory.
         app_dir = self.app_dir
 
         self._populate_staging(
-            name=name, version=version, clean=clean_staging
+            name=name, version=version, public_url=public_url,
+            clean=clean_staging
         )
 
         staging = pjoin(app_dir, 'staging')
@@ -379,7 +436,7 @@ class _AppHandler(object):
 
         messages = self.build_check(fast=True)
         if messages:
-            logger.info('\nBuild recommended:')
+            logger.info('\nBuild recommended, please run `jupyter lab build`:')
             [logger.info('    %s' % item) for item in messages]
 
     def build_check(self, fast=False):
@@ -397,9 +454,7 @@ class _AppHandler(object):
         if not osp.exists(pkg_path):
             return ['No built application']
 
-        with open(pkg_path) as fid:
-            static_data = json.load(fid)
-
+        static_data = self.info['static_data']
         old_jlab = static_data['jupyterlab']
         old_deps = static_data.get('dependencies', dict())
 
@@ -419,12 +474,12 @@ class _AppHandler(object):
             # Extensions that were added.
             for ext in new_jlab[ext_type]:
                 if ext not in old_jlab[ext_type]:
-                    messages.append('%s needs to be included' % ext)
+                    messages.append('%s needs to be included in build' % ext)
 
             # Extensions that were removed.
             for ext in old_jlab[ext_type]:
                 if ext not in new_jlab[ext_type]:
-                    messages.append('%s needs to be removed' % ext)
+                    messages.append('%s needs to be removed from build' % ext)
 
         # Look for mismatched dependencies
         for (pkg, dep) in new_deps.items():
@@ -514,6 +569,8 @@ class _AppHandler(object):
         linked[info['name']] = info['source']
         self._write_build_config(config)
 
+        return True
+
     def unlink_package(self, path):
         """Link a package by name or at the given path.
         """
@@ -554,6 +611,60 @@ class _AppHandler(object):
             disabled.remove(extension)
         self._write_page_config(config)
 
+    def check_extension(self, extension, check_installed_only=False):
+        """Check if a lab extension is enabled or disabled
+        """
+        info = self.info
+
+        if extension in info["core_extensions"]:
+            return self._check_core_extension(
+                extension, info, check_installed_only)
+
+        if extension in info["linked_packages"]:
+            self.logger.info('%s:%s' % (extension, GREEN_ENABLED))
+            return True
+
+        return self._check_common_extension(
+            extension, info, check_installed_only)
+
+    def _check_core_extension(self, extension, info, check_installed_only):
+        """Check if a core extension is enabled or disabled
+        """
+        if extension in info['uninstalled_core']:
+            self.logger.info('%s:%s' % (extension, RED_X))
+            return False
+        if check_installed_only:
+            self.logger.info('%s: %s' % (extension, GREEN_OK))
+            return True
+        if extension in info['disabled_core']:
+            self.logger.info('%s: %s' % (extension, RED_DISABLED))
+            return False
+        self.logger.info('%s:%s' % (extension, GREEN_ENABLED))
+        return True
+
+    def _check_common_extension(self, extension, info, check_installed_only):
+        """Check if a common (non-core) extension is enabled or disabled
+        """
+        if extension not in info['extensions']:
+            self.logger.info('%s:%s' % (extension, RED_X))
+            return False
+
+        errors = self._get_extension_compat()[extension]
+        if errors:
+            self.logger.info('%s:%s (compatibility errors)' % (extension, RED_X))
+            return False
+
+        if check_installed_only:
+            self.logger.info('%s: %s' % (extension, GREEN_OK))
+            return True
+
+        if _is_disabled(extension, info['disabled']):
+            self.logger.info('%s: %s' % (extension, RED_DISABLED))
+            return False
+
+        self.logger.info('%s:%s' % (extension, GREEN_ENABLED))
+        return True
+
     def _get_app_info(self):
         """Get information about the app.
         """
@@ -575,7 +686,12 @@ class _AppHandler(object):
                 sys.append(name)
 
         info['uninstalled_core'] = self._get_uninstalled_core_extensions()
-        info['version'] = core_data['jupyterlab']['version']
+
+        info['static_data'] = _get_static_data(self.app_dir)
+        app_data = info['static_data'] or core_data
+        info['version'] = app_data['jupyterlab']['version']
+        info['publicUrl'] = app_data['jupyterlab'].get('publicUrl', '')
+
         info['sys_dir'] = self.sys_dir
         info['app_dir'] = self.app_dir
 
@@ -589,7 +705,8 @@ class _AppHandler(object):
         info['disabled_core'] = disabled_core
         return info
 
-    def _populate_staging(self, name=None, version=None, clean=False):
+    def _populate_staging(self, name=None, version=None, public_url=None,
+            clean=False):
         """Set up the assets in the staging directory.
         """
         app_dir = self.app_dir
@@ -616,10 +733,19 @@ class _AppHandler(object):
                 overwrite_lock = False
 
         for fname in ['index.js', 'webpack.config.js',
+                'webpack.prod.config.js',
                 'yarn.lock', '.yarnrc', 'yarn.js']:
-            if fname == 'yarn.lock' and not overwrite_lock:
+            target = pjoin(staging, fname)
+            if (fname == 'yarn.lock' and os.path.exists(target) and
+                    not overwrite_lock):
                 continue
-            shutil.copy(pjoin(HERE, 'staging', fname), pjoin(staging, fname))
+            shutil.copy(pjoin(HERE, 'staging', fname), target)
+
+        # Ensure a clean templates directory
+        templates = pjoin(staging, 'templates')
+        if osp.exists(templates):
+            shutil.rmtree(templates)
+        shutil.copytree(pjoin(HERE, 'staging', 'templates'), templates)
 
         # Ensure a clean linked packages directory.
         linked_dir = pjoin(staging, 'linked_packages')
@@ -630,11 +756,24 @@ class _AppHandler(object):
         # Template the package.json file.
         # Update the local extensions.
         extensions = self.info['extensions']
+        removed = False
         for (key, source) in self.info['local_extensions'].items():
+            # Handle a local extension that was removed.
+            if key not in extensions:
+                config = self._read_build_config()
+                data = config.setdefault('local_extensions', dict())
+                del data[key]
+                self._write_build_config(config)
+                removed = True
+                continue
             dname = pjoin(app_dir, 'extensions')
             self._update_local(key, source, dname, extensions[key],
                 'local_extensions')
 
+        # Update the list of local extensions if any were removed.
+        if removed:
+            self.info['local_extensions'] = self._get_local_extensions()
+
         # Update the linked packages.
         linked = self.info['linked_packages']
         for (key, item) in linked.items():
@@ -651,6 +790,9 @@ class _AppHandler(object):
         if name:
             data['jupyterlab']['name'] = name
 
+        if public_url:
+            data['jupyterlab']['publicUrl'] = public_url
+
         pkg_path = pjoin(staging, 'package.json')
         with open(pkg_path, 'w') as fid:
             json.dump(data, fid, indent=4)
@@ -1064,17 +1206,23 @@ def _validate_extension(data):
 
     files = data['jupyterlab_extracted_files']
     main = data.get('main', 'index.js')
+    if not main.endswith('.js'):
+        main += '.js'
 
     if extension is True:
-        if main not in files:
-            messages.append('Missing extension module "%s"' % main)
-    elif extension and extension not in files:
-        messages.append('Missing extension module "%s"' % extension)
+        extension = main
+    elif extension and not extension.endswith('.js'):
+        extension += '.js'
 
     if mime_extension is True:
-        if main not in files:
-            messages.append('Missing mimeExtension module "%s"' % main)
-    elif mime_extension and mime_extension not in files:
+        mime_extension = main
+    elif mime_extension and not mime_extension.endswith('.js'):
+        mime_extension += '.js'
+
+    if extension and extension not in files:
+        messages.append('Missing extension module "%s"' % extension)
+
+    if mime_extension and mime_extension not in files:
         messages.append('Missing mimeExtension module "%s"' % mime_extension)
 
     if themeDir and not any(f.startswith(themeDir) for f in files):
@@ -1112,6 +1260,17 @@ def _get_core_data():
         return json.load(fid)
 
 
+def _get_static_data(app_dir):
+    """Get the data for the app static dir.
+    """
+    target = pjoin(app_dir, 'static', 'package.json')
+    if os.path.exists(target):
+        with open(target) as fid:
+            return json.load(fid)
+    else:
+        return None
+
+
 def _validate_compatibility(extension, deps, core_data):
     """Validate the compatibility of an extension.
     """
@@ -1222,3 +1381,7 @@ def _get_core_extensions():
     """
     data = _get_core_data()['jupyterlab']
     return list(data['extensions']) + list(data['mimeExtensions'])
+
+
+if __name__ == '__main__':
+    watch_dev(HERE)
